'''
[Fast Neural Style Transfer]
Justin Johnson, et al. Perceptual Losses for Real-Time Style Transfer and Super-Resolution. ECCV 2016.
Code reference: https://github.com/dxyang/StyleTransfer

[运行]
python Practice_25_FastNeuralStyleTransfer

[环境要求]
- Pytorch
- torchvision
- opencv
- matplotlib
- pillow
- tqdm

[模型]
在本文件加的checkpoint目录下下载以下模型
wget -c --no-check-certificate https://bethgelab.org/media/uploads/pytorch_models/vgg_conv.pth

[数据]
在本文件夹的data/styletransfer/目录下下载以下数据用作测试
wget -O ./data/styletransfer/mosaic.jpg "https://raw.githubusercontent.com/dxyang/StyleTransfer/master/style_imgs/mosaic.jpg"
wget -O ./data/styletransfer/amber.jpg "https://raw.githubusercontent.com/dxyang/StyleTransfer/master/content_imgs/amber.jpg"

在本文件夹的data/coco目录下下载以下数据用作训练
wget http://images.cocodataset.org/zips/train2017.zip --no-check-certificate
unzip -q train2017.zip
rm -f train2017.zip
然后将文件结构整理如下
data/coco/
└── train2017
    ├── 000000000009.jpg
    ├── ...
    └── 000000581929.jpg
'''

import os
from PIL import Image
import matplotlib.pyplot as plt

import torchvision.transforms as transforms
import torchvision.models as models

import numpy as np
from datetime import datetime
import time
import logging
import random
import cv2
import math

import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
import torch.autograd as autograd
from torch.autograd import Variable
from torch.nn import Linear, Conv2d, BatchNorm2d, LayerNorm, LeakyReLU, ConvTranspose2d, ReLU, Tanh, InstanceNorm2d
from torch.nn import ReflectionPad2d, ReplicationPad2d

from torch.utils.data import DataLoader
import torchvision.datasets as dset
import torchvision
import itertools
from tqdm import tqdm


'''辅助函数''' 
# 保存图像
def save_image(img, filename):
    tmp = ((img.detach().numpy().transpose(1, 2, 0) + 1.0) * 127.5).astype(np.uint8)
    cv2.imwrite(filename, cv2.cvtColor(tmp, cv2.COLOR_RGB2BGR))
    
# 图像的预处理和后处理函数
img_size = 256 # 训练图像全部resize到256*256
prep = transforms.Compose([transforms.Resize(img_size),
                           transforms.CenterCrop(img_size),
                           transforms.ToTensor(),
                           transforms.Lambda(lambda x: x[torch.LongTensor([2,1,0])]), 
                           transforms.Normalize(mean=[0.40760392, 0.45795686, 0.48501961], 
                                                std=[1,1,1]),
                           transforms.Lambda(lambda x: x.mul_(255)),
                          ])
prep_test = transforms.Compose([transforms.Resize(512),
                           transforms.CenterCrop(512), 
                           transforms.ToTensor(),
                           transforms.Lambda(lambda x: x[torch.LongTensor([2,1,0])]), 
                           transforms.Normalize(mean=[0.40760392, 0.45795686, 0.48501961], 
                                                std=[1,1,1]),
                           transforms.Lambda(lambda x: x.mul_(255)),
                          ])
postpa = transforms.Compose([transforms.Lambda(lambda x: x.mul_(1./255)),
                           transforms.Normalize(mean=[-0.40760392, -0.45795686, -0.48501961], 
                                                std=[1,1,1]),
                           transforms.Lambda(lambda x: x[torch.LongTensor([2,1,0])]), 
                           ])

# 将图像的范围限制到[-1,1]
def postp(tensor): 
    t = postpa(tensor) * 2 - 1
    t[t>1] = 1    
    t[t<-1] = -1
    return t


'''损失函数定义'''
# 计算Gram矩阵
class GramMatrix(nn.Module):
    def forward(self, input):
        b,c,h,w = input.size()
        F = input.view(b, c, h*w)
        G = torch.bmm(F, F.transpose(1,2)) 
        G.div_(h*w)
        return G

# 计算风格损失
# # input为VGG特征，target为目标风格特征，首先将input转换为Gram矩阵，再计算其与target的均方误差
class GramMSELoss(nn.Module):
    def forward(self, input, target):
        out = nn.MSELoss()(GramMatrix()(input), target)
        return(out)

# 定义风格特征的层数    
style_layers = ['r12','r22','r32','r43'] 
# 定义内容特征的层数
content_layers = ['r33']
# 整合所有计算损失函数的类
loss_layers = style_layers + content_layers
loss_fns = [GramMSELoss()] * len(style_layers) + [nn.MSELoss()] * len(content_layers)
if torch.cuda.is_available():
    loss_fns = [loss_fn.cuda() for loss_fn in loss_fns]
    
# 不同层风格损失和内容损失的权重
style_weights = [1e1/n**2 for n in [64,128,256,512]]
content_weights = [1e0]
weights = style_weights + content_weights

    
'''网络定义'''
# 定义VGG网络模型
class VGG(nn.Module):
    def __init__(self, pool='max'):
        super(VGG, self).__init__()
        #vgg modules
        self.conv1_1 = nn.Conv2d(3, 64, kernel_size=3, padding=1)
        self.conv1_2 = nn.Conv2d(64, 64, kernel_size=3, padding=1)
        self.conv2_1 = nn.Conv2d(64, 128, kernel_size=3, padding=1)
        self.conv2_2 = nn.Conv2d(128, 128, kernel_size=3, padding=1)
        self.conv3_1 = nn.Conv2d(128, 256, kernel_size=3, padding=1)
        self.conv3_2 = nn.Conv2d(256, 256, kernel_size=3, padding=1)
        self.conv3_3 = nn.Conv2d(256, 256, kernel_size=3, padding=1)
        self.conv3_4 = nn.Conv2d(256, 256, kernel_size=3, padding=1)
        self.conv4_1 = nn.Conv2d(256, 512, kernel_size=3, padding=1)
        self.conv4_2 = nn.Conv2d(512, 512, kernel_size=3, padding=1)
        self.conv4_3 = nn.Conv2d(512, 512, kernel_size=3, padding=1)
        self.conv4_4 = nn.Conv2d(512, 512, kernel_size=3, padding=1)
        self.conv5_1 = nn.Conv2d(512, 512, kernel_size=3, padding=1)
        self.conv5_2 = nn.Conv2d(512, 512, kernel_size=3, padding=1)
        self.conv5_3 = nn.Conv2d(512, 512, kernel_size=3, padding=1)
        self.conv5_4 = nn.Conv2d(512, 512, kernel_size=3, padding=1)
        if pool == 'max':
            self.pool1 = nn.MaxPool2d(kernel_size=2, stride=2)
            self.pool2 = nn.MaxPool2d(kernel_size=2, stride=2)
            self.pool3 = nn.MaxPool2d(kernel_size=2, stride=2)
            self.pool4 = nn.MaxPool2d(kernel_size=2, stride=2)
            self.pool5 = nn.MaxPool2d(kernel_size=2, stride=2)
        elif pool == 'avg':
            self.pool1 = nn.AvgPool2d(kernel_size=2, stride=2)
            self.pool2 = nn.AvgPool2d(kernel_size=2, stride=2)
            self.pool3 = nn.AvgPool2d(kernel_size=2, stride=2)
            self.pool4 = nn.AvgPool2d(kernel_size=2, stride=2)
            self.pool5 = nn.AvgPool2d(kernel_size=2, stride=2)
            
    # out_keys为指定的层数，根据out_keys提取x经过VGG后对应层数的特征        
    def forward(self, x, out_keys):
        out = {}
        out['r11'] = F.relu(self.conv1_1(x))
        out['r12'] = F.relu(self.conv1_2(out['r11']))
        out['p1'] = self.pool1(out['r12'])
        out['r21'] = F.relu(self.conv2_1(out['p1']))
        out['r22'] = F.relu(self.conv2_2(out['r21']))
        out['p2'] = self.pool2(out['r22'])
        out['r31'] = F.relu(self.conv3_1(out['p2']))
        out['r32'] = F.relu(self.conv3_2(out['r31']))
        out['r33'] = F.relu(self.conv3_3(out['r32']))
        out['r34'] = F.relu(self.conv3_4(out['r33']))
        out['p3'] = self.pool3(out['r34'])
        out['r41'] = F.relu(self.conv4_1(out['p3']))
        out['r42'] = F.relu(self.conv4_2(out['r41']))
        out['r43'] = F.relu(self.conv4_3(out['r42']))
        out['r44'] = F.relu(self.conv4_4(out['r43']))
        out['p4'] = self.pool4(out['r44'])
        out['r51'] = F.relu(self.conv5_1(out['p4']))
        out['r52'] = F.relu(self.conv5_2(out['r51']))
        out['r53'] = F.relu(self.conv5_3(out['r52']))
        out['r54'] = F.relu(self.conv5_4(out['r53']))
        out['p5'] = self.pool5(out['r54'])
        return [out[key] for key in out_keys]

# 卷积模块，包含卷积层+实例归一化层+激活层
class ConvLayer(nn.Module):
    def __init__(self, in_channels, out_channels, kernel_size, stride):
        super(ConvLayer, self).__init__()
        padding = kernel_size // 2
        self.pad = nn.ReflectionPad2d(padding)
        self.conv = nn.Conv2d(in_channels, out_channels, kernel_size, stride)
        self.relu = nn.ReLU()
        self.IN = nn.InstanceNorm2d(out_channels, affine=True)

    def forward(self, x):
        return self.relu(self.IN(self.conv(self. pad(x))))

# 反卷积模块，包含卷积层+实例归一化层+激活层
class UpsampleConvLayer(nn.Module):
    def __init__(self, in_channels, out_channels, kernel_size, stride):
        super(UpsampleConvLayer, self).__init__()
        self.upsample = nn.Upsample(scale_factor=2, mode='nearest')
        reflection_padding = kernel_size // 2
        self.pad = nn.ReflectionPad2d(reflection_padding)
        self.conv = nn.Conv2d(in_channels, out_channels, kernel_size, stride)
        self.relu = nn.ReLU()
        self.IN = nn.InstanceNorm2d(out_channels, affine=True)

    def forward(self, x):
        return self.relu(self.IN(self.conv(self. pad(self.upsample(x)))))

# 残差模块，输入的特征经过两个卷积层后与自身相加
class ResidualBlock(nn.Module):
    def __init__(self, channels):
        super(ResidualBlock, self).__init__()
        self.conv1 = ConvLayer(channels, channels, kernel_size=3, stride=1)
        self.IN1 = nn.InstanceNorm2d(channels, affine=True)
        self.relu = nn.ReLU()
        self.conv2 = ConvLayer(channels, channels, kernel_size=3, stride=1)
        self.IN2 = nn.InstanceNorm2d(channels, affine=True)

    def forward(self, x):
        out = self.IN2(self.conv2(self.relu(self.IN1(self.conv1(x)))))
        out = out + x
        return self.relu(out)

class ImageTransformNet(nn.Module):
    def __init__(self):
        super(ImageTransformNet, self).__init__()

        modelList = []       
        
        # 编码器包含3层卷积层，特征被下采2次
        modelList.append(ConvLayer(3, 32, kernel_size=9, stride=1))
        modelList.append(ConvLayer(32, 64, kernel_size=3, stride=2))
        modelList.append(ConvLayer(64, 128, kernel_size=3, stride=2))

        # 变换模块包含5层残差模块
        for n in range(5):
            modelList.append(ResidualBlock(128))

        # 解码器包含3层卷积层，特征被上采2次
        modelList.append(UpsampleConvLayer(128, 64, kernel_size=3, stride=1))
        modelList.append(UpsampleConvLayer(64, 32, kernel_size=3, stride=1))
        modelList.append(nn.ReflectionPad2d(9//2))
        modelList.append(nn.Conv2d(32, 3, kernel_size=9, stride=1))

        self.model = nn.Sequential(*modelList)

    def forward(self, x):
        return self.model(x)       
    

'''构建，加载和初始化模型'''
# 构建VGG网络实例，并加载参数vgg_conv.pth
vgg = VGG()
vgg.load_state_dict(torch.load('./checkpoint/vgg_conv.pth'))
# 不训练网络，因此网络不需要计算梯度
for param in vgg.parameters():
    param.requires_grad = False
if torch.cuda.is_available():
    vgg.cuda()
print(vgg)

# 图像转换网络
image_transformer = ImageTransformNet()
if torch.cuda.is_available():
    image_transformer.cuda()
print(image_transformer)

    
'''数据准备''' 
image_dir = './data/styletransfer/'
img_dirs = [image_dir, image_dir]
# 通过改变下面的文件名更改训练用的风格图像和测试用的内容图像
img_names = ['mosaic.jpg', 'amber.jpg']
imgs = [Image.open(img_dirs[i] + name) for i,name in enumerate(img_names)]
imgs_torch = [prep(imgs[0]), prep_test(imgs[1])]
if torch.cuda.is_available():
    imgs_torch = [Variable(img.unsqueeze(0).cuda()) for img in imgs_torch]
else:
    imgs_torch = [Variable(img.unsqueeze(0)) for img in imgs_torch]
# 风格图像和内容图像分别为style_image和content_image
style_image, content_image = imgs_torch

# 图像style_image的风格特征
style_targets = [GramMatrix()(A).detach() for A in vgg(style_image, style_layers)]

# 训练集
train_dataset = dset.ImageFolder('./data/coco/', prep)
train_loader = DataLoader(train_dataset, batch_size = 4)
print('Dataset size:', len(train_loader.dataset.imgs))


'''网络训练'''
tv_weight = 1e-7 # 全变分损失的权重 
# 优化器更新image_transformer网络参数而不是图像 
optimizer = torch.optim.Adam(image_transformer.parameters(), 1e-3) 

pbar = tqdm(range(40000), initial=0, smoothing=0.01, dynamic_ncols=True)

iterator = iter(train_loader)
for idx in pbar:
    try:
        data, _ = next(iterator)
    except StopIteration:
        iterator = iter(dataloader)
        data, _ = next(iterator) 
    x = data.cuda()
    # 计算目标内容特征，而目标风格特征固定不变可以提前计算好 
    content_targets = [A.detach() for A in vgg(x, content_layers)] 
    targets = style_targets + content_targets 
    y_hat = image_transformer(x) 
    # 下面这部分与神经图像风格化一致 
    out = vgg(y_hat, loss_layers) 
    layer_losses = [weights[a] * loss_fns[a](A, targets[a]) for a,A in enumerate(out)] 
    # 计算全变分损失，利用1个偏移来计算水平和竖直方向的相邻像素差 
    diff_i = torch.sum(torch.abs(y_hat[:, :, :, 1:] - y_hat[:, :, :, :-1])) 
    diff_j = torch.sum(torch.abs(y_hat[:, :, 1:, :] - y_hat[:, :, :-1, :])) 
    layer_losses += [tv_weight*(diff_i + diff_j)] 
    loss = sum(layer_losses) 
    
    optimizer.zero_grad() # 将网络的参数梯度初始化为0 
    loss.backward() # 计算梯度并反向传播 
    optimizer.step() # 更新网络的参数    
    
    if idx % 1000 == 0:
        with torch.no_grad():
            y_hat0 = image_transformer(content_image) 
        out_img = postp(y_hat0.data[0].cpu().squeeze())
        save_image(out_img, 'fastNST_%d.jpg'%idx)
    
    pbar.set_description(
        (
            f"content_loss: {layer_losses[-2].data.mean().item():.3f}; "
            f"style_loss: {sum(layer_losses[:-2]).data.mean().item():.3f}; "
            f"tv_loss: {layer_losses[-1].data.mean().item():.3f};"
        )
    )

# 保存模型    
torch.save(image_transformer.state_dict(),  f"fastNST-%05d.pt"%(idx))

print('Done!')